/**
 * @ddumbfs.c
 * @author  Alain Spineux <alain.spineux@gmail.com>
 * @version 0.0
 *
 * @section LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details at
 * http://www.gnu.org/copyleft/gpl.html
 *
 * @section DESCRIPTION
 *
 * ddumbfs.c provides the FUSE interface for the ddumbfs filesystem
 *
 */

#define _XOPEN_SOURCE 500
#define _LARGEFILE64_SOURCE

#define FUSE_USE_VERSION 26
//#define  HAVE_SETXATTR

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GNU_SOURCE

#include <fuse.h>
#include <ulockmgr.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/time.h>
#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif

#include <stddef.h>
#include <inttypes.h>

#include <ctype.h>

#include <pthread.h>

#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <assert.h>

#include <ftw.h>
#include <search.h>


#define NO_ULTRAFAST

#ifdef ULTRAFAST
    // disable all mutex and spin lock to evaluate the overhead
    #define pthread_spin_lock(x) 0
    #define pthread_spin_unlock(x) 0
    #define pthread_mutex_lock(x) 0
    #define pthread_mutex_unlock(x) 0
    #define pthread_cond_signal(x) {}
    #define pthread_cond_broadcast(x) {}
    #define pthread_cond_wait(x,y) {}
    #define pthread_rwlock_rdlock(x,y) {}
    #define pthread_rwlock_wrlock(x,y) {}
    #define pthread_rwlock_unlock(x,y) {}
#endif

#define XSTAT_FH_SZ      256  // at least 12 to backup VMware ESX via NFS, because vmmkfstools open 8 simultaneous connection to the same file

#include "bits.h"
#include "ddfslib.h"
#include "ddfschkrep.h"

pthread_mutex_t bfile_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_spinlock_t ifile_spinlock;

long long int r_file_count;
long long int r_addr_count;
long long int r_frag_count;
long long int r_not_found;

// cannot start reclaim() when in ddumb_block_write(), this require a pthread_cond_t and its mutex
pthread_mutex_t reclaim_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t reclaim_ddumb_block_write_is_unused=PTHREAD_COND_INITIALIZER;
int reclaim_ddumb_block_write_is_in_use=0;

// I need to know when reclaim() is running and maintain a list of recently allocated block
pthread_spinlock_t reclaim_spinlock;
struct bit_array ba_found_in_files;
int reclaim_enable=0;

// handle save of block list at regulare interval
long long int used_block_saved=-1;
long long int used_block_current=0;
int ddumbfs_background_terminate=0;
pthread_t ddumbfs_background_pthread;
pthread_mutex_t ddumb_background_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t ddumb_background_cond=PTHREAD_COND_INITIALIZER;


typedef struct ddumb_param
{
    char  *parent;
    int   pool;
    int   check_at_start;
    int   lock_index;
    int   direct_io;
} struct_ddumb_param;

struct_ddumb_param ddumb_param = { NULL, 0, 0, 1, 1 };

struct ddumb_fh;

// simultaneous access to a file are protected by zones
struct xzone
{
    long long int start;
    long long int end;
    char op;
    char right;
};

// xstat is for extra stat (currently only file size)
struct xstat
{


    long long int ino;      // only one xstat per ino

    struct file_header h;   // the payload
    int saved;              // does the payload need to be saved ?

    int ref;                             // number of ddumb_fh pointing to this xstat, mostly for debugging
    struct ddumb_fh *fhs[XSTAT_FH_SZ];   // ddumb_fh pointing to this xstat, used to search for ddumb_fh->buf
    int fhs_n;
    pthread_mutex_t xstat_lock;
    pthread_cond_t  zone_cond;
    pthread_cond_t  buf_cond;

};

#define DDFS_BUF_EMPTY  0
#define DDFS_BUF_RDONLY 1
#define DDFS_BUF_RDWR   2

// this is the info that are attached to fuse struct fuse_file_info *fi
struct ddumb_fh
{   // one is created by open, the same can be used simultaneously by multiple thread
    int fd;
    int rdonly;
    struct xstat *xstat;
    char *filename;         // the filename (could be NULL)
    int pool_writer;        // number of "writer" thread busy on this fh
    char *buf;              // attached buffer, used only for write operation
    int buf_loaded;         // do this buffer contains relevant data that must be saved
    long long int buf_off;  // the offset of this buffer inside the file
    pthread_mutex_t lock;
    int special;
    int delayed_write_error_code;
    struct xzone zone;
};

// maintains a tree sorted by ino with all currently open files
struct xstat_root
{
    void *root;
    pthread_mutex_t mutex;
    struct xstat* xstat;
} xstat_root;


struct ddumb_statistic
{   // contains some usefull stat about the running ddumbfs
    long int header_load;
    long int header_save;

    long int hash;

    long int block_write;
    long int read_before_write; // write not on a block boundary, requiring a read
    long int eof_write;         // write after eof
    long int ghost_write;       // block already exist, just reuse the block address
    long int block_write_try_next_node; // node already used, try next
    long int block_write_slide;  // slide inside node block
    long int block_read;
    long int block_read_zero;    // read zero block, not from disk
    long int write_save;          // the write is not sequential, save an "uncomplete" buffer

    long int getattr;
    long int fgetattr;
    long int do_truncate;
    long int ftruncate;
    long int truncate;
    long int create;
    long int open;
    long int read;
    long int write;
    long int flush;
    long int release;
    long int fsync;
    long int fh_counter;        // number of ddumb_fh simultaneously open

    long int reclaim;
} ddumb_statistic;

// early declaration
static int _ddumb_write(const char *path, const char *buf, size_t size, off_t offset, struct ddumb_fh *fh);

#define WRITE_FIELD(file, field) { if (ddumb_statistic.field>0) fprintf(file, "%-30s %9ld\n", #field, ddumb_statistic.field); }

static void ddumb_write_statistic(FILE *file)
{   // write ddumbfs statistics to FILE *file
    WRITE_FIELD(file, header_load);
    WRITE_FIELD(file, header_save);

    WRITE_FIELD(file, hash);

    WRITE_FIELD(file, block_write);
    WRITE_FIELD(file, read_before_write);
    WRITE_FIELD(file, ghost_write);
    WRITE_FIELD(file, write_save);
    WRITE_FIELD(file, eof_write);

    WRITE_FIELD(file, block_write_try_next_node);
    WRITE_FIELD(file, block_write_slide);

    WRITE_FIELD(file, block_read);
    WRITE_FIELD(file, block_read_zero);

    WRITE_FIELD(file, getattr);
    WRITE_FIELD(file, fgetattr);
    WRITE_FIELD(file, do_truncate);
    WRITE_FIELD(file, ftruncate);
    WRITE_FIELD(file, truncate);
    WRITE_FIELD(file, create);
    WRITE_FIELD(file, open);
    WRITE_FIELD(file, read);
    WRITE_FIELD(file, write);
    WRITE_FIELD(file, flush);
    WRITE_FIELD(file, release);
    WRITE_FIELD(file, fsync);

    WRITE_FIELD(file, reclaim);

    long long int s, u;
    pthread_spin_lock(&ifile_spinlock);
    bit_array_count(&ddfs.ba_usedblocks, &s, &u); // no need to lock ddfs.ba_usedblocks here
    pthread_spin_unlock(&ifile_spinlock);
    fprintf(file, "%-30s %9lld\n", "block_allocated", s);
    fprintf(file, "%-30s %9lld\n", "block_free", u);
    fprintf(file, "%-30s %9.2f\n", "overflow", ddfs.c_node_overflow);
    fprintf(file, "%-30s %9d\n", "direct_io", ddfs.direct_io);
    fprintf(file, "%-30s %9d\n", "lock_index", ddfs.lock_index);
    fprintf(file, "%-30s %9s\n", "hash", ddfs.c_hash);
}

unsigned long int thread_id(void)
{
    return (unsigned long int) pthread_self();
}

void preload_node(long long int node_idx)
{  // make a read in nodes to load required page, usually just before to lock the mutex
    char _ch=*(ddfs.nodes+(node_idx*ddfs.c_node_size));
    _ch=_ch;
}

long long int min_int(long long int a, long long int b)
{
    if (a<b) return a;
    return b;
}

struct ddumb_fh *ddumb_get_fh(struct fuse_file_info *fi)
{
    return (struct ddumb_fh *)(uintptr_t)fi->fh;
}

/*
 * xstat_*
 */
static int xstat_load(int fd, struct xstat *xstat, const char* filename)
{   // load extra stat data from file header
    ddumb_statistic.header_load++;

    int len=file_header_set(fd, &xstat->h);
    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "xstat_load pread: %s (%s)\n", filename, strerror(errno));
        return -errno;
    }
    else if (len==0)
    {
        DDFS_LOG(LOG_ERR, "xstat_load, not a valid ddumbfs file: %s\n", filename);
        return -EIO;
    }
    else if (len!=ddfs.c_file_header_size)
    {
        DDFS_LOG(LOG_ERR, "xstat_load read only %d/%d bytes: %s\n", len, ddfs.c_file_header_size, filename);
        return -EIO;
    }

    xstat->saved=1;
    DDFS_LOG_DEBUG("[%lu]++  xstat_load fd=%d len=%d xstat=%p size=%llx(%lld) %s\n", thread_id(), fd , len, (void*)xstat, (long long int)xstat->h.size, (long long int)xstat->h.size, filename);
    return 0;
}

//#define pthread_mutex_lock_d(x)   { DDFS_LOG_DEBUG("[%lu]++ pthread_mutex_lock   %p\n", thread_id(), (void*) (x)); pthread_mutex_lock(x); }
//#define pthread_mutex_unlock_d(x) { DDFS_LOG_DEBUG("[%lu]++ pthread_mutex_unlock %p\n", thread_id(), (void*) (x)); pthread_mutex_unlock(x); }

#define pthread_mutex_lock_d(x)   { pthread_mutex_lock(x); }
#define pthread_mutex_unlock_d(x) { pthread_mutex_unlock(x); }

static int xstat_save(int fd, struct xstat *xstat, const char* filename)
{   // save extra stat data into file header
    ddumb_statistic.header_save++;
    DDFS_LOG_DEBUG("[%lu]++  xstat_save fd=%d size=%lld %s\n", thread_id(), fd, (long long int)xstat->h.size, filename);

    int len=file_header_get(fd, &xstat->h);
    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "xstat_save pwrite: %s (%s)\n", filename, strerror(errno));
        return -errno;
    }
    else if (len!=ddfs.c_file_header_size)
    {
        DDFS_LOG(LOG_ERR, "xstat_save wrote only %d/%d bytes: %s\n", len, ddfs.c_file_header_size, filename);
        return -EIO;
    }
    xstat->saved=1;
    return 0;
}

static int xstat_save_fh(struct ddumb_fh *fh)
{   // save extra stat data into file header
    DDFS_LOG_DEBUG("[%lu]++  xstat_save_fi fd=%d fh=%p rdonly=%d %s\n", thread_id(), fh->fd, (void*)fh, fh->rdonly, fh->filename);
    if (!fh->rdonly) return xstat_save(fh->fd, fh->xstat, fh->filename);
    return 0;
}

int xstat_compare(const void *a, const void* b)
{
    return ((struct xstat *)a)->ino - ((struct xstat *)b)->ino;
}

int _xstat_register(struct xstat_root *root, struct ddumb_fh *fh)
{
    int res=0;

//    DDFS_LOG_DEBUG("[%lu]++  xstat_register fd=%d fh=%p root->xstat=%p\n", thread_id(), fh->fd, fh, (void*)root->xstat);
    if (!root->xstat)
    {
        root->xstat=malloc(sizeof(struct xstat));
        if (root->xstat==NULL)
        {
            DDFS_LOG(LOG_ERR, "xstat_register: xstat malloc failed: %s\n", fh->filename);
            return -ENOMEM;
        }
    }

    struct stat stbuf;
    if (fstat(fh->fd, &stbuf)==-1)
    {
        DDFS_LOG(LOG_ERR, "xstat_register: cannot fstat the file: %s (%s)\n", fh->filename, strerror(errno));
        return -errno;
    }
    root->xstat->ino=stbuf.st_ino;

    void *val=tsearch((const void *)root->xstat, &(root->root), xstat_compare);
//    DDFS_LOG_DEBUG("[%lu]++  tsearch root=%p xtstat=%p  -> xstat=%p\n", thread_id(), (void *)root->root, (void *)root->xstat, (void*)*(struct xstat**)val);
    if (val==NULL)
    {
        DDFS_LOG(LOG_ERR, "xstat_register: tsearch cannot allocate a new item: %s\n", fh->filename);
        fh->xstat=NULL;
        return -ENOMEM;
    }

    fh->xstat=*(struct xstat**)val;
    if (fh->xstat==root->xstat)
    {
        root->xstat=NULL;
//        DDFS_LOG_DEBUG("[%lu]** xstat_register xstat=%p memeset size=%d\n\n", thread_id(), fh->xstat, sizeof(fh->xstat->fhs));
        fh->xstat->ref=1;
        fh->xstat->fhs[0]=fh;
        fh->xstat->fhs_n=1;
        DDFS_LOG_DEBUG("[%lu]pthread_mutex_init xstat %p %s\n", thread_id(), (void*)&fh->xstat->xstat_lock, fh->filename);
        pthread_mutex_init(&fh->xstat->xstat_lock, 0);
        pthread_cond_init(&fh->xstat->zone_cond, NULL);
        pthread_cond_init(&fh->xstat->buf_cond, NULL);

        res=xstat_load(fh->fd, fh->xstat, fh->filename);
        if (res<0)
        {
            tdelete(&(fh->xstat), &(root->root), xstat_compare);
            root->xstat=fh->xstat;
            fh->xstat=NULL;
        }
    }
    else
    {
        fh->xstat->ref++;
        pthread_mutex_lock_d(&fh->xstat->xstat_lock);
        // add fh to fh->xstat->fhs
        assert(fh->xstat->fhs_n<XSTAT_FH_SZ); // fh->xstat->fhs is not full
        fh->xstat->fhs[fh->xstat->fhs_n++]=fh;
        pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
    }
    DDFS_LOG_DEBUG("[%lu]++  xstat_register fd=%d fh=%p xstat=%p ref=%d root->xstat=%p %s\n", thread_id(), fh->fd, (void*)fh, (void*)fh->xstat, fh->xstat->ref, (void*)root->xstat, fh->filename);
    return res;
}

int xstat_register(struct xstat_root *root, struct ddumb_fh *fh)
{
    int res;
    pthread_mutex_lock(&(root->mutex));
    res=_xstat_register(root, fh);
    pthread_mutex_unlock(&(root->mutex));
    return 0;
}

void xstat_subscribe(struct xstat_root *root, struct ddumb_fh *fh_src, struct ddumb_fh *fh_dst)
{
    pthread_mutex_lock(&(root->mutex));
    fh_src->xstat->ref++; // protected by root->mutex
    char *buf=fh_dst->buf;
    fh_dst->fd=fh_src->fd;
    fh_dst->buf=fh_src->buf;
    fh_dst->buf_off=fh_src->buf_off;

    pthread_mutex_lock_d(&fh_src->xstat->xstat_lock);
    fh_src->buf=buf;
    // add fh_dst to fh_src->xstat->fhs
    fh_dst->xstat=fh_src->xstat;
    assert(fh_src->xstat->fhs_n<XSTAT_FH_SZ); // fh_src->xstat->fhs is not full
    fh_src->xstat->fhs[fh_src->xstat->fhs_n++]=fh_dst;

    fh_dst->buf_loaded=DDFS_BUF_RDONLY;
    fh_src->buf_loaded=DDFS_BUF_EMPTY;
    fh_dst->zone.start=fh_dst->buf_off;
    fh_dst->zone.end=fh_dst->buf_off+ddfs.c_block_size;
    // fh_dst->zone.op='F';     // zone of writers are set at creation never change
    // fh_dst->zone.right='r';  // zone of writers are set at creation never change

    pthread_mutex_unlock_d(&fh_src->xstat->xstat_lock);
    pthread_mutex_unlock(&(root->mutex));
}

void xstat_unsubscribe(struct xstat_root *root, struct ddumb_fh *fh)
{
    pthread_mutex_lock(&(root->mutex));
    fh->xstat->ref--;
    assert(fh->xstat->ref>0);

    int i, found;
    pthread_mutex_lock_d(&fh->xstat->xstat_lock);
    // fh->zone.op='N';    // zone of writers are set at creation never change
    // fh->zone.right='n'; // zone of writers are set at creation never change
    for(i=0, found=0; i<fh->xstat->fhs_n; i++)
    {
        found=(fh->xstat->fhs[i]==fh);
        if (found)
        {   // move last entry into the free one
            fh->xstat->fhs_n--;
            if (i<fh->xstat->fhs_n) fh->xstat->fhs[i]=fh->xstat->fhs[fh->xstat->fhs_n];
            break;
        }
    }
    pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
    assert(found);
    pthread_mutex_unlock(&(root->mutex));
}

int xstat_get(struct xstat_root *root, struct xstat *xstat, const char *path)
{
    // get xstat from tsearch tree or from header file if not found
    // idem xstat_load but don't need xstat_release, use tfind instead of tsearch, this is read only
    int res=0;
    pthread_mutex_lock(&(root->mutex));

    void *val=tfind((const void *)xstat, &(root->root), xstat_compare);
//  DDFS_LOG_DEBUG("[%lu]++  xstat_get tsearch root=%p xtstat=%p  -> xstat=%p\n", thread_id(), (void *)root->root, (void *)root->xstat, (void*)*(struct xstat**)val);
    if (val==NULL)
    {
        int fd=open(path, O_RDONLY);
        if (fd==-1)
        {
            DDFS_LOG(LOG_ERR, "xstat_get: cannot open %s to read header: %s\n", path, strerror(errno));
            res=-errno;
        }
        else
        {
            res=xstat_load(fd, xstat, path);
            close(fd);
        }
    }
    else
    {
        file_header_copy(&xstat->h, &(*(struct xstat**)val)->h);
    }
    pthread_mutex_unlock(&(root->mutex));
    return res;
}

int xstat_release(struct xstat_root *root, struct ddumb_fh *fh)
{
    int res=0;
    void *p;

    DDFS_LOG_DEBUG("[%lu]++  xstat_release fd=%d fh=%p xstat=%p root->xstat=%p %s\n", thread_id(), fh->fd, (void*)fh, (void*)fh->xstat, (void*)root->xstat, fh->filename);

    pthread_mutex_lock(&(root->mutex));

    fh->xstat->ref--;
    assert(fh->xstat->ref>=0);

    if (fh->xstat->ref==0)
    {
        // cannot update header if file was open readonly
        if (!fh->rdonly && !fh->xstat->saved) res=xstat_save(fh->fd, fh->xstat, fh->filename);
        p=tdelete(fh->xstat, &(root->root), xstat_compare);
        if (p==NULL) DDFS_LOG(LOG_ERR, "xstat_release: xstat not found, continue: %s\n", fh->filename);

#ifdef DDFS_DEBUG
        // double check if one and only one is still equal to fh
        assert(fh->xstat->fhs_n==1 && fh->xstat->fhs[0]==fh);
        // double check if all zone are free
        assert(fh->xstat->xzones_n==0);
#endif

        pthread_mutex_destroy(&fh->xstat->xstat_lock);
        pthread_cond_destroy(&fh->xstat->zone_cond);
        pthread_cond_destroy(&fh->xstat->buf_cond);

        // reuse this xstat pointer if possible, else free it
        if (root->xstat==NULL) root->xstat=fh->xstat;
        else
        {
            free(fh->xstat);
            fh->xstat=NULL;
        }
    }
    else
    {
        int i, found;
        pthread_mutex_lock_d(&fh->xstat->xstat_lock);
        for(i=0; i<fh->xstat->fhs_n; i++)
        {
            if (fh->xstat->fhs[i]==fh)
            {   // move last entry into the free one
                found=1;
                fh->xstat->fhs_n--;
                if (i<fh->xstat->fhs_n) fh->xstat->fhs[i]=fh->xstat->fhs[fh->xstat->fhs_n];
                break;
            }
        }
        pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
        assert(found);
    }

    pthread_mutex_unlock(&(root->mutex));
    return res;
}

static int zone_overlap(long long int offset1, long long int size1, long long int offset2, long long int size2)
{   // never used/tested !!!!
    return ((offset1<=offset2 && offset2<offset1+size1) ||
            (offset2<=offset1 && offset1<offset2+size2));
}

static void xzone_lock(struct ddumb_fh *fh, long long int offset, long long int size, char op)
{   // lock a zone
    struct xstat *xstat=fh->xstat;
    char right='n';

    pthread_mutex_lock_d(&xstat->xstat_lock);
    fh->zone.op=op;

    int busy=1; // zone is already in use
    while (busy)
    {
        if (op=='F')
        {   // ddumb_buffer_flush
            // not yet used
            assert(0); // ddumb_buffer_flush use xstat_subscribe() instead, this code is never used
            fh->zone.start=offset;
            fh->zone.end=offset+size;
            right='r';
        }
        else if (op=='R')
        {   // read
            fh->zone.start=ddfs.c_block_size*(offset/ddfs.c_block_size);
            fh->zone.end=ddfs.c_block_size*((offset+size+ddfs.c_block_size-1)/ddfs.c_block_size);
            right='r';
        }
        else if (op=='W')
        {   // write
            if (offset<xstat->h.size) fh->zone.start=ddfs.c_block_size*(offset/ddfs.c_block_size);
            else fh->zone.start=ddfs.c_block_size*(xstat->h.size/ddfs.c_block_size);
            if (offset+size>xstat->h.size) fh->zone.end=-1;
            else fh->zone.end=ddfs.c_block_size*((offset+size+ddfs.c_block_size-1)/ddfs.c_block_size);
            right='w';
        }
        else if (op=='T')
        {   // truncate
            if (offset<xstat->h.size) fh->zone.start=ddfs.c_block_size*(offset/ddfs.c_block_size);
            else fh->zone.start=ddfs.c_block_size*(xstat->h.size/ddfs.c_block_size);
            fh->zone.end=-1;
            right='w';
        }

        if (right!='f')
        {
            // right=='f' NOT YET USED // always true, ddumb_buffer_flush use xstat_subscribe() instead
            int i;
            for(i=0, busy=0; !busy && i<fh->xstat->fhs_n; i++)
            {
                struct ddumb_fh *xfh=fh->xstat->fhs[i];
                if (xfh->zone.right=='n') continue;

                busy=((xfh->zone.start<=fh->zone.start && (fh->zone.start<xfh->zone.end || xfh->zone.end==-1)) ||
                      (fh->zone.start<=xfh->zone.start && (xfh->zone.start<xfh->zone.end || fh->zone.end==-1))) &&
                      (right!='r' || xfh->zone.right!='r');
            }
            if (busy)
            {
                DDFS_LOG_DEBUG("[%lu]ZONE WAIT op=%c fd=%d ino=%lld start=0x%llx(%lld) end=0x%llx(%lld) fh=%p xfh=%p xop=%c xstart=0x%llx xend=0x%llx %s\n",
                        thread_id(), fh->zone.op, fh->fd, (long long int)fh->xstat->ino, fh->zone.start, fh->zone.start, fh->zone.end, fh->zone.end, fh, xfh, xfh->zone.op, xfh->zone.start, xfh->zone.end, fh->filename);
                pthread_cond_wait(&fh->xstat->zone_cond, &fh->xstat->xstat_lock);
            }
        }
    }
    fh->zone.right=right;
    pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
    DDFS_LOG_DEBUG("[%lu]ZONE LOCK fh=%p op=%c fd=%d ino=%lld start=0x%llx(%lld) end=0x%llx(%lld) %s\n", thread_id(), fh, fh->zone.op, fh->fd, (long long int)fh->xstat->ino, fh->zone.start, fh->zone.start, fh->zone.end, fh->zone.end, fh->filename);
}

static void xzone_unlock(struct ddumb_fh *fh)
{   // unlock a zone
    pthread_mutex_lock_d(&fh->xstat->xstat_lock);
    DDFS_LOG_DEBUG("[%lu]ZONE UNLOCK fh=%p op=%c ino=%lld start=0x%llx(%lld) end=0x%llx(%lld) %s\n", thread_id(), fh, fh->zone.op, (long long int)fh->xstat->ino, fh->zone.start, fh->zone.start, fh->zone.end, fh->zone.end, fh->filename);
    fh->zone.op='N';
    fh->zone.right='n';
    pthread_cond_broadcast(&fh->xstat->zone_cond);
    pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
}


/*
 * ddumb_*_fh
 */
struct ddumb_fh *ddumb_alloc_fh(struct fuse_file_info *fi, int fd, int rdonly, int special, const char* filename)
{   // alloc a new ddumb_fh and attache it to the fi
    int res=0;

    struct ddumb_fh *fh=malloc(sizeof(struct ddumb_fh));
    if (fh)
    {
        fh->fd=fd;
        fh->buf_loaded=DDFS_BUF_EMPTY;
        fh->buf_off=-1;
        fh->pool_writer=0;
        fh->rdonly=rdonly;
        fh->special=special;
        fh->buf=NULL;
        fh->xstat=NULL;
        fh->delayed_write_error_code=0;
        fh->zone.op='N';
        fh->zone.right='n';
        fh->filename=strdup(filename); // don't care about the failure, just for friendly error message
        if (!special)
        {
            if (ddfs.direct_io)
            {
                res=posix_memalign((void *)&fh->buf, BLOCK_ALIGMENT, ddfs.c_block_size);
                if (res)
                {
                    DDFS_LOG(LOG_ERR, "ddumb_alloc_fh: cannot allocate aligned buffer: %s (%s)\n", filename, strerror(res));
                }
            }
            else
            {
                fh->buf=malloc(ddfs.c_block_size);
            }

            if (fd!=-1)
            {   // not attached to the writer_pool
                if (fh->buf) res=xstat_register(&xstat_root, fh); // FIXME: The error must be reported as a MEMORY ERROR
                else DDFS_LOG(LOG_ERR, "ddumb_alloc_fh: buffer malloc failed: %s\n", filename);
            }

            if (fh->buf==NULL || (fd!=-1 && fh->xstat==NULL) || res<0)
            {
                free(fh->buf);
                if (fh->filename) free(fh->filename);
                free(fh);
                fh=NULL;
            }
            else
            {
                ddumb_statistic.fh_counter++;
                pthread_mutex_init(&fh->lock, 0);
                DDFS_LOG_DEBUG("[%lu]pthread_mutex_init lock %p\n", thread_id(), (void*)&fh->lock);
            }
        }
    }
    else DDFS_LOG(LOG_ERR, "ddumb_alloc_fh: malloc failed: %s\n", filename);

    fi->fh=(uintptr_t)fh;
    return fh;
}

static int ddumb_free_fh(struct fuse_file_info *fi)
{   // free the ddumb_fh attached to the fi
    int res=0;

    struct ddumb_fh *fh=ddumb_get_fh(fi);
    if (fh)
    {
        if (fh->filename) free(fh->filename);
        if (!fh->special)
        {
            res=xstat_release(&xstat_root, fh);
            free(fh->buf);
            fh->buf=NULL;
            pthread_mutex_destroy(&fh->lock);
            free(fh);
            ddumb_statistic.fh_counter--;
        }
    }
    fi->fh=(unsigned long)NULL;
    return res;
}


int tree_explore(const char *fpath, const struct stat *sb, int typeflag, struct FTW *ftwbuf)
{   // called by nftw to walk the tree
    int res, len;
    unsigned char node[ADDR_SIZE];
    unsigned char *hash=node+ddfs.c_addr_size;
    long long int addr;
    long long int prev_addr=-1;

    if (typeflag!=FTW_F) return 0;
    if (0==strncmp(fpath+ddfs.rdir_len, SPECIAL_DIR, ddfs.special_dir_len))
    {   // skip SPECIAL_DIR
        return 0;
    }

    r_file_count++;

    FILE *file=fopen(fpath, "r");  // FILE* are faster here
    if (file==NULL)
    {
        DDFS_LOG(LOG_ERR, "reclaim cannot open: %s (%s)\n", fpath, strerror(errno));
        return 1;
    }

    // be careful, file_header_set_conv return c_file_header_size for empty file
    uint64_t file_size;
    len=file_header_set_conv(fileno(file), &file_size);
    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "reclaim cannot read header: %s (%s)\n", fpath, strerror(errno));
        fclose(file);
        return 1;
    }
    else if (len!=ddfs.c_file_header_size)
    {
        DDFS_LOG(LOG_ERR, "reclaim invalid ddumbfs header, skip file: %s\n", fpath);
        fclose(file);
        return 0;
    }

    res=fseek(file, ddfs.c_file_header_size, SEEK_SET);
    if (res==-1)
    {
        DDFS_LOG(LOG_ERR, "reclaim cannot seek after header: %s (%s)\n", fpath, strerror(errno));
        fclose(file);
        return 1;
    }

    len=fread(node, 1, ddfs.c_node_size, file);
    while (len==ddfs.c_node_size)
    {
        addr=ddfs_get_node_addr(node);

        if (prev_addr!=-1 && prev_addr+1!=addr && addr!=0) r_frag_count++;
        if (addr!=0) prev_addr=addr;

        if (addr>ddfs.c_block_count)
        {
            DDFS_LOG(LOG_WARNING, "reclaim invalid block address %lld: %s\n", addr, fpath);
        }
        else if (addr!=0)
        {
            // fprintf(stderr,"addr=%lld\n", addr);
            r_addr_count++;

            // search the index, if the same hash can be found for another block
            // this will avoid to maybe loose useful information for next fsck
            blockaddr baddr=addr;
            pthread_spin_lock(&ifile_spinlock);
            nodeidx node_idx=ddfs_search_hash(hash, &baddr);
            if (node_idx<0) r_not_found++;
            pthread_spin_unlock(&ifile_spinlock);

            pthread_spin_lock(&reclaim_spinlock);
            bit_array_set(&ba_found_in_files, addr);
            // if another block was found, keep it too. I don't want to loose information
            // that could be verified and maybe useful for fsck.
            if (baddr!=addr) bit_array_set(&ba_found_in_files, baddr);
            pthread_spin_unlock(&reclaim_spinlock);
        }
        len=fread(node, 1, ddfs.c_node_size, file);
    }

    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "reclaim cannot read: %s (%s)\n", fpath, strerror(errno));
        fclose(file);
        return 1;
    }

    if (len && len!=ddfs.c_node_size)
    {
        DDFS_LOG(LOG_WARNING, "reclaim read incomplete node (%d/%d): %s\n", len, ddfs.c_node_size, fpath);
    }
    fclose(file);
    return 0;
}

int ddumbfs_save_usedblocks(int limit)
{   // if required, save used block list
    int res=0;
    long long int _u;

    if (limit==0)
    {
        bit_array_count(&ddfs.ba_usedblocks, &used_block_current, &_u);
    }

//    DDFS_LOG_DEBUG("save used block %lld-%lld>%d\n", used_block_current, used_block_saved, limit);
    if (llabs(used_block_current-used_block_saved)>=limit)
    {
        res=ddfs_save_usedblocks();
        if (res==0)
        {
            used_block_saved=used_block_current;
            DDFS_LOG(LOG_INFO, "save used block list: %lld blocks used\n", used_block_current);
        }
    }
    return res;
}

void reclaim(FILE *output)
{   // the reclaim() process
    int res;
    int success=0;

    DDFS_LOG(LOG_INFO, "reclaim started (online).\n");
    L_SYS(LOG_INFO, "online reclaim started on: %s\n", ddfs.pdir);

    // paranoid test
    char *specialdir=SPECIAL_DIR;
    if (!isdir(specialdir+1))
    {
        DDFS_LOG(LOG_ERR, "system directory %s not found in root!\n", SPECIAL_DIR);
        return;
    }

    // reclaim() must wait for everybody to have left ddumb_block_write
    // before to enter this part (up to unlock reclaim_mutex)
    pthread_mutex_lock(&reclaim_mutex);
    while (reclaim_ddumb_block_write_is_in_use>0) pthread_cond_wait(&reclaim_ddumb_block_write_is_unused, &reclaim_mutex);
    reclaim_ddumb_block_write_is_in_use++; // don't enter in reclaim() when it is already running

    pthread_spin_lock(&reclaim_spinlock);
    reclaim_enable=1;
    bit_array_reset(&ba_found_in_files, 0);
    bit_array_set(&ba_found_in_files, 0); // 0 is reserved
    bit_array_set(&ba_found_in_files, 1); // 1 is reserved
    pthread_spin_unlock(&reclaim_spinlock);

    pthread_mutex_unlock(&reclaim_mutex);

    r_file_count=r_addr_count=r_frag_count=r_not_found=0;
    long long int start=now();
    res=nftw(ddfs.rdir, tree_explore, 10, FTW_MOUNT | FTW_PHYS);
    long long int end=now();

    if (res==-1)
    {
        DDFS_LOG(LOG_ERR, "reclaim error, cannot run, correct and check filesystem\n");
        L_SYS(LOG_ERR, "reclaim error, cannot run, correct and check filesystem\n");
        if (output) fprintf(output, "error: reclaim cannot run, correct errors and check the filesystem\n");
    }
    else
    {
        long long int block_allocated, block_not_allocated;
        long long int block_in_use, block_not_in_use;
        DDFS_LOG(LOG_INFO, "reclaim read used blocks from %lld files in %.1fs\n", r_file_count, (end-start)*1.0/NOW_PER_SEC);
        if (output) fprintf(output, "== Read used blocks from files in %.1fs\n", (end-start)*1.0/NOW_PER_SEC);

        int cmp, cmp_res;

        pthread_spin_lock(&ifile_spinlock);
        pthread_spin_lock(&reclaim_spinlock);
        cmp_res=bit_array_cmp(&ba_found_in_files, &ddfs.ba_usedblocks, &cmp);
        bit_array_count(&ddfs.ba_usedblocks, &block_allocated, &block_not_allocated);
        bit_array_count(&ba_found_in_files, &block_in_use, &block_not_in_use);
        pthread_spin_unlock(&reclaim_spinlock);
        pthread_spin_unlock(&ifile_spinlock);

        if (!cmp_res || cmp>0)
        {
            DDFS_LOG(LOG_ERR, "reclaim inconsistency between allocated and used blocks, filesystem check required\n");
            L_SYS(LOG_ERR, "reclaim inconsistency between allocated and used blocks, filesystem check required\n");
            if (output) fprintf(output, "Critical error: inconsistency between allocated and used blocks, filesystem check required\n");
        }
        else
        {
            long long int r_addr_count_nz=r_addr_count;
            if (r_addr_count_nz==0) r_addr_count_nz=1;

            DDFS_LOG(LOG_INFO, "reclaim   block_allocated=%9lld  block_not_allocated=%9lld\n", block_allocated, block_not_allocated);
            DDFS_LOG(LOG_INFO, "reclaim      block_in_use=%9lld        block_to_free=%9lld\n", block_in_use, block_allocated-block_in_use);
            DDFS_LOG(LOG_INFO, "reclaim    hash_not_found=%9lld                files=%9lld\n", r_not_found, r_file_count);
            DDFS_LOG(LOG_INFO ,"reclaim block_references=%10lld        fragmentation=%8.1f%%\n", r_addr_count, r_frag_count*100.0/r_addr_count_nz);
            if (output) fprintf(output, "%-30s %9lld\n", "block_allocated", block_allocated);
            if (output) fprintf(output, "%-30s %9lld\n", "block_not_allocated", block_not_allocated);
            if (output) fprintf(output, "%-30s %9lld\n", "block_in_use", block_in_use);
            if (output) fprintf(output, "%-30s %9lld\n", "block_not_in_use", block_not_in_use);
            if (output) fprintf(output, "%-30s %9lld\n", "hash_not_found", r_not_found);
            if (output) fprintf(output, "%-30s %9lld\n", "files", r_file_count);
            if (output) fprintf(output, "%-30s %9lld\n", "block_references", r_addr_count);
            if (output) fprintf(output, "%-30s %8.1f%%\n", "fragmentation", r_frag_count*100.0/r_addr_count_nz);

            start=now();

            long long int node_count=0;
            long long int node_deleted=0;
            long long int node_idx=0;
            // read all nodes, delete unused blocks
            while (node_idx<ddfs.c_node_count)
            {
                preload_node(node_idx);  // pre-load into cache before to lock
                pthread_spin_lock(&ifile_spinlock);
                long long int addr=ddfs_get_node_addr(ddfs.nodes+(node_idx*ddfs.c_node_size));
                if (addr==0)
                {
                    node_idx++;
                }
                else
                {
                    pthread_spin_lock(&reclaim_spinlock);
                    int found=bit_array_get(&ba_found_in_files, addr);
                    pthread_spin_unlock(&reclaim_spinlock);
                    if (found)
                    {
                        node_count++;
                        node_idx++;
                    }
                    else
                    {
                        node_delete(node_idx);
                        bit_array_unset(&ddfs.ba_usedblocks, addr);
                        node_deleted++;
                    }
                }
                // don't increment node_idx, because nodes have moved up and current node is a new node
                pthread_spin_unlock(&ifile_spinlock);
            }
            end=now();
            success=1;
            ddumb_statistic.reclaim++;
            if (output) fprintf(output, "== Index cleanup in %.1fs\n", (end-start)*1.0/NOW_PER_SEC);
            L_SYS(LOG_INFO, "reclaim     node_in_index=%9lld         node_deleted=%9lld\n", node_count, node_deleted);
            DDFS_LOG(LOG_INFO, "reclaim     node_in_index=%9lld         node_deleted=%9lld\n", node_count, node_deleted);
            if (output) fprintf(output, "%-30s %9lld\n", "nodes_in_index", node_count);
            if (output) fprintf(output, "%-30s %9lld\n", "node_deleted", node_deleted);
        }
    }
    pthread_spin_lock(&reclaim_spinlock);
    reclaim_enable=0;
    pthread_spin_unlock(&reclaim_spinlock);

    pthread_spin_lock(&ifile_spinlock);
    ddumbfs_save_usedblocks(0);
    if (success && ddfs.c_reuse_asap)
    {
        ddfs.ba_usedblocks.index=DDFS_LAST_RESERVED_BLOCK+1; // reuse recently freed blocks ASAP
    }
    pthread_spin_unlock(&ifile_spinlock);

    pthread_mutex_lock(&reclaim_mutex);
    reclaim_ddumb_block_write_is_in_use--; // you can call reclaim() again
    pthread_mutex_unlock(&reclaim_mutex);
}

static int ddumb_simple_block_read(struct ddumb_fh *fh, char *buf, long long int offset, long long int size)
{
    unsigned char baddr[ADDR_SIZE];
    long long int gap=offset%ddfs.c_block_size;

    DDFS_LOG_DEBUG("[%lu]++  ddumb_simple_block_read offset=0x%llx(%lld) size=0x%llx(%lld) gap=%lld %s\n", thread_id(), offset, offset, size, size, gap, fh->filename);

    // read the address of the block in the block file
    long long int idx_off=offset/ddfs.c_block_size*ddfs.c_node_size+ddfs.c_file_header_size;

    int len=pread(fh->fd, baddr, ddfs.c_addr_size, idx_off);
    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "ddumb_simple_block_read pread %s (%s)\n", fh->filename, strerror(errno));
        return -errno;
    }
    else if (len==0)
    {
        DDFS_LOG(LOG_ERR, "ddumb_simple_block_read read 0 bytes, maybe end of file. file_size=%lld offset=%lld %s\n", fh->xstat->h.size, offset, fh->filename);
        return -EIO;
    }
    else if (len!=ddfs.c_addr_size)
    {
        DDFS_LOG(LOG_ERR, "ddumb_simple_block_read partial read, only %d/%d bytes %s\n", len, ddfs.c_addr_size, fh->filename);
        return -EIO;
    }

    long long int block_addr=ddfs_get_node_addr(baddr);

    if (block_addr==0)
    {   // addr==0 means block '\0......\0'
        ddumb_statistic.block_read_zero++;
        DDFS_LOG_DEBUG("[%lu]-- ddumb_simple_block_read read ZERO block fd=%d fh=%p\n", thread_id(), fh->fd, (void*)fh);
    }

    // TODO: This is a bottleneck, only one reader at a time, be careful for race condition with writer in ddfs_write_block2()
    pthread_mutex_lock(&bfile_mutex);
    len=ddfs_read_block(block_addr, buf, size, gap);
    pthread_mutex_unlock(&bfile_mutex);

    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "ddumb_simple_block_read reading blockfile off=%lld: %s\n", block_addr*ddfs.c_block_size+gap, strerror(errno));
        return -errno;
    }
    else if (len!=size)
    {
        DDFS_LOG(LOG_ERR, "ddumb_simple_block_read reading off=%lld, only %d/%lld bytes\n", block_addr*ddfs.c_block_size+gap, len, size);
        return -EIO;
    }
    return 0;
}

static int ddumb_block_read(struct ddumb_fh *fh, char *buf, long long int offset, long long int size)
{
    long long int gap=offset%ddfs.c_block_size;
    long long int block_boundary=offset-gap;

    ddumb_statistic.block_read++;

    DDFS_LOG_DEBUG("[%lu]++  ddumb_block_read fd=%d offset=0x%llx(%lld) size=0x%llx(%lld) gap=%lld fhs_n=%d %s\n", thread_id(), fh->fd, offset, offset, size, size, gap, fh->xstat->fhs_n, fh->filename);

    // search if the required block is already somewhere in one open fh
    int i, found;
    pthread_mutex_lock_d(&fh->xstat->xstat_lock);
    for(i=0, found=0; !found && i<fh->xstat->fhs_n; i++)
    {
        struct ddumb_fh *xfh=fh->xstat->fhs[i];
//        DDFS_LOG_DEBUG("[%lu]>>>>ddumb_block_read fh=%p xfh=%p buf_loaded=%d buf_off=0x%llx(%lld) I=%d \n", thread_id(), (void*)fh, (void*)xfh, xfh->buf_loaded, xfh->buf_off, xfh->buf_off, i);
        if (xfh->buf_loaded && xfh->buf_off==block_boundary)
        {
            found=1;
            memcpy(buf, xfh->buf+gap, size);
            DDFS_LOG_DEBUG("[%lu]--  ddumb_block_read fd=%d fh=%p FOUND in buffer xstat=%p xfd=%d xfh=%p I=%d %s\n", thread_id(), fh->fd, (void*)fh, (void*)fh->xstat, xfh->fd, (void*)xfh, i, fh->filename);
        }
    }
    pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
    if (found) return 0;

    return ddumb_simple_block_read(fh, buf, offset, size);
}

/**
 * write a block in the filesystem
 *
 * identical to ddfs_write_block, but handle statistics and multi-threading stuff
 *
 * @param block the block
 * @param bhash return the hash of the block
 * @return the address of the block or <0 for error
 */
long long int ddfs_write_block2(const char *block, unsigned char *bhash)
{
    long long int addr;
    long long int node_idx, node_idx0;

    ddfs_hash(block, bhash);
    ddumb_statistic.hash++;
    if (memcmp(bhash, ddfs.zero_block_hash, ddfs.c_hash_size)==0) return 0;

    // preload the page where the node is before to spinlock
    // node_idx is calculated twice, but....
    node_idx0=ddfs_hash2idx(bhash);
    preload_node(node_idx0);

    pthread_spin_lock(&ifile_spinlock);
    int res=ddfs_locate_hash(bhash, &addr, &node_idx);

    if (res<0)
    {
        pthread_spin_unlock(&ifile_spinlock);
        return res;
    }

    if (res==0)
    {   // the block is already stored, return its current address
        pthread_spin_unlock(&ifile_spinlock);
        ddumb_statistic.ghost_write++;
        return addr;
    }

    // update statistic
    ddumb_statistic.block_write_try_next_node+=node_idx-node_idx0;

    // the hash is not found
    // store the block in the BlockFile
    long long int baddr=ddfs_alloc_block();
    if (baddr<0)
    {
        pthread_spin_unlock(&ifile_spinlock);
        DDFS_LOG(LOG_ERR, "no more free block, start reclaim() procedure !\n");
        return baddr;
    }

    // save the used block list at regular interval
    used_block_current++;
    ddumbfs_save_usedblocks(16000);

    // I'm using this block, don't free it if reclaim is running
    // this is the first place where ba_found_in_files is updated
    pthread_spin_lock(&reclaim_spinlock);
    if (reclaim_enable && baddr!=-1) bit_array_set(&ba_found_in_files, baddr);
    pthread_spin_unlock(&reclaim_spinlock);

    if (addr!=0)
    {   // we must insert the node, search for a free node
        long long int free_node_idx=ddfs_search_free_node(node_idx+1, ddfs.c_node_count);
        if (free_node_idx<0)
        {   // this should NEVER NERVER NEVER append
            pthread_spin_unlock(&ifile_spinlock);
            return -EIO;
        }
        memmove(ddfs.nodes+((node_idx+1)*ddfs.c_node_size), ddfs.nodes+(node_idx*ddfs.c_node_size), (free_node_idx-node_idx)*ddfs.c_node_size);
        ddumb_statistic.block_write_slide++;
    }
    // now node_idx is ready to receive new node
    ddfs_set_node(node_idx, baddr, bhash);

    pthread_mutex_lock(&bfile_mutex);  // lock block file before to release ifile to avoid race condition
    pthread_spin_unlock(&ifile_spinlock);
    // now you can read the index, but not yet the blockfile because data are not yet in sync with the index
    baddr=ddfs_store_block(block, baddr);
    pthread_mutex_unlock(&bfile_mutex);

    return baddr;
}


static int ddumb_buf_write(struct ddumb_fh *fh)
{
    unsigned char node[ADDR_SIZE+HASH_SIZE];

    unsigned char *baddr=node;
    unsigned char *bhash=node+ddfs.c_addr_size;

    int len;
    int ret=0;

    // I cannot enter this function when reclaim() is _starting_
    // and reclaim() must wait for everybody to have left this function
    // before to start
    pthread_mutex_lock(&reclaim_mutex);
    reclaim_ddumb_block_write_is_in_use++;
    pthread_mutex_unlock(&reclaim_mutex);

    ddumb_statistic.block_write++;

    long long int addr=ddfs_write_block2(fh->buf, bhash);
    if (addr<0)
    {
        ret=addr;
    }
    else
    {
        // addr is already registered into ba_found_in_files by index_new_block

        DDFS_LOG_DEBUG("[%lu]++  ddumb_buf_write fd=%d offset=0x%llx(%lld) addr=%lld %s\n", thread_id(), fh->fd, fh->buf_off, fh->buf_off, addr, fh->filename);

        ddfs_convert_addr(addr, baddr);

        off_t addr_off=fh->buf_off/ddfs.c_block_size*ddfs.c_node_size+ddfs.c_file_header_size;

        len=pwrite(fh->fd, baddr, ddfs.c_node_size, addr_off);

        pthread_spin_lock(&reclaim_spinlock);
        // This is 2nd place where ba_found_in_files is updated
        if (reclaim_enable) bit_array_set(&ba_found_in_files, addr);
        pthread_spin_unlock(&reclaim_spinlock);

        if (len==-1)
        {
            DDFS_LOG(LOG_ERR, "ddumb_buf_write addr offset=%lld %s (%s)\n", (long long int)addr_off, fh->filename, strerror(errno));
            ret=-errno;
        }
        else if (len!=ddfs.c_node_size)
        {
            DDFS_LOG(LOG_ERR, "ddumb_buf_write addr offset=%lld wrote only %d/%d %s\n", (long long int)addr_off, len, ddfs.c_node_size, fh->filename);
            ret=-EIO;
        }
    }

    pthread_mutex_lock_d(&fh->xstat->xstat_lock);
    // buffer has been written (or not) and don't contain anything useful now
    fh->buf_loaded=DDFS_BUF_EMPTY;
    // warn everybody about the change
    pthread_cond_broadcast(&fh->xstat->buf_cond);
    pthread_mutex_unlock_d(&fh->xstat->xstat_lock);

    pthread_mutex_lock(&reclaim_mutex);
    reclaim_ddumb_block_write_is_in_use--;
    if (reclaim_ddumb_block_write_is_in_use==0) pthread_cond_signal(&reclaim_ddumb_block_write_is_unused);
    pthread_mutex_unlock(&reclaim_mutex);

    return ret;
}

struct ddumb_fh *writer_pool_loader=NULL;
pthread_mutex_t writer_pool_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t writer_pool_cond_accept=PTHREAD_COND_INITIALIZER; // loader is empty, accepting new data
pthread_cond_t writer_pool_cond_push=PTHREAD_COND_INITIALIZER;   // push new data into loader
pthread_cond_t writer_pool_cond_submit=PTHREAD_COND_INITIALIZER; // submit loader to a writer
pthread_cond_t writer_pool_cond_done=PTHREAD_COND_INITIALIZER;   // block has been written
pthread_t *writer_pool_threads=NULL;

static void *writer_pool_loop(void *ptr)
{
    struct ddumb_fh *fh=(struct ddumb_fh *)ptr;
    printf("writer started: %s\n", fh->filename);
    pthread_mutex_lock(&writer_pool_mutex);
    while (1)
    {
        struct ddumb_fh *fh_src;

        // Wait for new data to handle in writer_pool_loader
        while (writer_pool_loader==NULL)
        {
            DDFS_LOG_DEBUG("[%lu]**  writer_pool_loop wait for a push of new data %s\n", thread_id(), fh->filename);
            pthread_cond_wait(&writer_pool_cond_push, &writer_pool_mutex);
        }

        fh_src=writer_pool_loader;
        writer_pool_loader=NULL;
        DDFS_LOG_DEBUG("[%lu]**  writer_pool_loop writer_pool_loader set to NULL\n", thread_id());
        // loader is ready to accept new data for other writers of the pool
        pthread_cond_signal(&writer_pool_cond_accept);
        pthread_mutex_unlock(&writer_pool_mutex);

        xstat_subscribe(&xstat_root, fh_src, fh);

        DDFS_LOG_DEBUG("[%lu]**  writer_pool_loop got fh=%p fd=%d offset=0x%llx(%lld) %s\n", thread_id(), fh_src, fh->fd, fh->buf_off, fh->buf_off, fh->filename);

        // fh has been submitted, signal the sender
        pthread_mutex_lock(&writer_pool_mutex);
        pthread_cond_broadcast(&writer_pool_cond_submit);
        pthread_mutex_unlock(&writer_pool_mutex);

        // write block
        // DDFS_LOG_DEBUG("[%lu]**  writer_pool_loop NULL fh_src=%p fd=%d offset=0x%llx(%lld) data=0x%llx %s\n", thread_id(), fh_src, fh->fd, fh->buf_off, fh->buf_off, *(long long int*)fh->buf, fh->filename);
        fh_src->delayed_write_error_code=ddumb_buf_write(fh);

        xstat_unsubscribe(&xstat_root, fh);

        pthread_mutex_lock(&writer_pool_mutex);
        fh_src->pool_writer--;
        assert(fh_src->pool_writer>=0);
        pthread_cond_broadcast(&writer_pool_cond_done);
        DDFS_LOG_DEBUG("[%lu]**  writer_pool_loop DONE fh_src=%p writer=%d fd=%d offset=0x%llx(%lld) data=0x%llx %s\n", thread_id(), fh_src, fh_src->pool_writer, fh->fd, fh->buf_off, fh->buf_off, *(long long int*)fh->buf, fh->filename);
    }
    pthread_mutex_unlock(&writer_pool_mutex);
    return NULL;
}

static int writer_pool_handle(struct ddumb_fh *fh)
{
    assert(!fh->rdonly);
    assert(!fh->special);

    DDFS_LOG_DEBUG("[%lu]++  writer_pool_handle fh=%p writer=%d fd=%d offset=0x%llx(%lld) delayed_err=%d %s\n", thread_id(), fh, fh->pool_writer, fh->fd, fh->buf_off, fh->buf_off, fh->delayed_write_error_code, fh->filename);
    if (fh->delayed_write_error_code) return fh->delayed_write_error_code;
    pthread_mutex_lock(&writer_pool_mutex);
    // wait for writer_pool_loader to be empty to push fh in it
    while (writer_pool_loader!=NULL)
    {
        DDFS_LOG_DEBUG("[%lu]    writer_pool_handle wait for loader to accept fh=%p writer=%d fd=%d offset=0x%llx(%lld) %s\n", thread_id(), fh, fh->pool_writer, fh->fd, fh->buf_off, fh->buf_off, fh->filename);
        pthread_cond_wait(&writer_pool_cond_accept, &writer_pool_mutex);
    }
    // push fh in writer_pool_loader
    writer_pool_loader=fh;
    fh->pool_writer++;
    // signal the writer_pool for data in writer_pool_loader
    pthread_cond_signal(&writer_pool_cond_push);
    DDFS_LOG_DEBUG("[%lu]    writer_pool_handle pushed fh=%p writer=%d fd=%d offset=0x%llx(%lld) %s\n", thread_id(), fh, fh->pool_writer, fh->fd, fh->buf_off, fh->buf_off, fh->filename);

    // wait for the block to be submited
    while (fh->buf_loaded!=DDFS_BUF_EMPTY) // I don't thing this is risky to access fh->buf_loaded without locking fh->xstat->xstat_lock
    {
        DDFS_LOG_DEBUG("[%lu]    writer_pool_handle wait to submit fh=%p writer=%d fd=%d offset=0x%llx(%lld) %s\n", thread_id(), fh, fh->pool_writer, fh->fd, fh->buf_off, fh->buf_off, fh->filename);
        pthread_cond_wait(&writer_pool_cond_submit, &writer_pool_mutex);
    }
    pthread_mutex_unlock(&writer_pool_mutex);
    DDFS_LOG_DEBUG("[%lu]--  writer_pool_handle submitted fh=%p writer=%d fd=%d offset=0x%llx(%lld) %s\n", thread_id(), fh, fh->pool_writer, fh->fd, fh->buf_off, fh->buf_off, fh->filename);
    return 0;
}

static int init_writer_pool()
{
    struct fuse_file_info fi;
    char thread_name[64];
    int i;

    writer_pool_loader=NULL;
    for (i=0; i<ddumb_param.pool; i++)
    {
        snprintf(thread_name, sizeof(thread_name), "writer-%d", i);
        struct ddumb_fh *fh=ddumb_alloc_fh(&fi, -1, 0, 0, thread_name);
        if (fh==NULL)
        {
            DDFS_LOG(LOG_ERR, "init_writer_pool not enough memory for ddumb_alloc_fh()\n");
            return -ENOMEM;
        }
        fh->zone.op='F';
        fh->zone.right='n';
        int ret=pthread_create(writer_pool_threads+i, NULL, &writer_pool_loop, fh);
        if (ret)
        {
            DDFS_LOG(LOG_ERR, "init_writer_pool pthread_create: (%s)\n", strerror(errno));
            return -errno;
        }
    }
    DDFS_LOG(LOG_INFO, "writer pool: %d threads\n", ddumb_param.pool);
    L_SYS(LOG_INFO, "writer pool: %d threads\n", ddumb_param.pool);
    return 0;
}

static int ddumb_buffer_flush(struct ddumb_fh *fh)
{
    int res=0;

    // I should not lock using a spin lock here, because we have to wait for
    // end of disk write, but nobody else should try to get the resource
    // except VMware backup in 2gbparse in some cases

    if (fh->buf_loaded)
    {
        assert(fh->buf_loaded!=DDFS_BUF_RDONLY); // don't flush the same buffer twice
        DDFS_LOG_DEBUG("[%lu]    ddumb_buffer_flush fd=%d offset=0x%llx(%lld) data=0x%llx %s\n", thread_id(), fh->fd, fh->buf_off, fh->buf_off, *(long long int*)fh->buf, fh->filename);
        if (ddumb_param.pool)
        {
            res=writer_pool_handle(fh);
        }
        else
        {
            pthread_mutex_lock_d(&fh->xstat->xstat_lock);
            fh->buf_loaded=DDFS_BUF_RDONLY; // will be flushed
            pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
            res=ddumb_buf_write(fh);
        }
    }
//    DDFS_LOG_DEBUG("[%lu]--  ddumb_buffer_flush fd=%d offset=0x%llx(%lld) data=0x%llx %s\n", thread_id(), fh->fd, fh->buf_off, fh->buf_off, *(long long int*)fh->buf, fh->filename);

    return res;
}

static int ddumb_getattr(const char *path, struct stat *stbuf)
{
    int res;

    ddumb_statistic.getattr++;
    int special=(0==strncmp(path, SPECIAL_DIR, ddfs.special_dir_len));

    if (strcmp(path, "/")==0) path="/.";

    res=lstat(path+1, stbuf);
    if (res==-1) return -errno;

    if (S_ISREG(stbuf->st_mode) && !special)
    {
        // retrieve size
        struct xstat xstat;
        xstat.ino=stbuf->st_ino;
        res=xstat_get(&xstat_root, &xstat, path+1);
        if (res>=0)
        {
            stbuf->st_size=xstat.h.size;
            stbuf->st_blksize=ddfs.c_block_size;
            stbuf->st_blocks=(xstat.h.size+ddfs.c_block_size-1)/ddfs.c_block_size*ddfs.c_block_size/512;
        }

        if (res<0) return res;
    }

    DDFS_LOG_DEBUG("[%lu]++  ddumb_getattr %s size=%lld\n", thread_id(), path, (long long int)stbuf->st_size);
    return 0;
}

static int ddumb_fgetattr(const char *path, struct stat *stbuf, struct fuse_file_info *fi)
{
    (void) path;
    int res;

    ddumb_statistic.fgetattr++;
    struct ddumb_fh *fh=ddumb_get_fh(fi);

    res=fstat(fh->fd, stbuf);
    if (res==-1) return -errno;

    if (!fh->special)
    {
        stbuf->st_size=fh->xstat->h.size;
        stbuf->st_blocks=(fh->xstat->h.size+ddfs.c_block_size-1)/ddfs.c_block_size*ddfs.c_block_size/512;
        stbuf->st_blksize=ddfs.c_block_size;
    }
//    DDFS_LOG_DEBUG("[%lu]++  ddumb_fgetattr size=%lld xstat=%p\n", thread_id(), (long long int)stbuf->st_size, (void*)fh->xstat);
    return 0;
}

static int ddumb_access(const char *path, int mask)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res = access(path+1, mask);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_readlink(const char *path, char *buf, size_t size)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res = readlink(path+1, buf, size - 1);
    if (res == -1)
        return -errno;

    buf[res] = '\0';
    return 0;
}

struct ddumb_dirp {
    DIR *dp;
    struct dirent *entry;
    off_t offset;
};

static int ddumb_opendir(const char *path, struct fuse_file_info *fi)
{
    int res;
    struct ddumb_dirp *d = malloc(sizeof(struct ddumb_dirp));
    if (d == NULL)
        return -ENOMEM;

    if (strcmp(path, "/")==0) path="/.";

    d->dp = opendir(path+1);
    if (d->dp == NULL) {
        res = -errno;
        free(d);
        return res;
    }
    d->offset = 0;
    d->entry = NULL;

    fi->fh = (unsigned long) d;
    return 0;
}

static inline struct ddumb_dirp *get_dirp(struct fuse_file_info *fi)
{
    return (struct ddumb_dirp *) (uintptr_t) fi->fh;
}

static int ddumb_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi)
{
    struct ddumb_dirp *d = get_dirp(fi);

    (void) path;
    if (offset != d->offset) {
        seekdir(d->dp, offset);
        d->entry = NULL;
        d->offset = offset;
    }
    while (1) {
        struct stat st;
        off_t nextoff;

        if (!d->entry) {
            d->entry = readdir(d->dp);
            if (!d->entry)
                break;
        }

        memset(&st, 0, sizeof(st));
        st.st_ino = d->entry->d_ino;
        st.st_mode = d->entry->d_type << 12;
        nextoff = telldir(d->dp);
        if (filler(buf, d->entry->d_name, &st, nextoff))
            break;

        d->entry = NULL;
        d->offset = nextoff;
    }

    return 0;
}

static int ddumb_releasedir(const char *path, struct fuse_file_info *fi)
{
    struct ddumb_dirp *d = get_dirp(fi);
    (void) path;
    closedir(d->dp);
    free(d);
    return 0;
}

static int ddumb_mknod(const char *path, mode_t mode, dev_t rdev)
{
    int res;

    DDFS_LOG_DEBUG("[%lu]++  ddumb_mknod %s mode=%o fifo=%d sock=%d\n", thread_id(), path, mode, mode&S_IFIFO, mode&S_IFSOCK);
    if (strcmp(path, "/")==0) path="/.";
    if (S_ISFIFO(mode))
        res = mkfifo(path+1, mode);
    else
        res = mknod(path+1, mode, rdev);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_mkdir(const char *path, mode_t mode)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res = mkdir(path+1, mode);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_unlink(const char *path)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res = unlink(path+1);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_rmdir(const char *path)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res = rmdir(path+1);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_symlink(const char *from, const char *to)
{
    int res;

    if (strcmp(to, "/")==0) to="/.";
    res = symlink(from, to+1);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_rename(const char *from, const char *to)
{
    int res;

    if (strcmp(from, "/")==0) from="/.";
    if (strcmp(to, "/")==0) to="/.";
    res = rename(from+1, to+1);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_link(const char *from, const char *to)
{
    int res;

    if (strcmp(from, "/")==0) from="/.";
    if (strcmp(to, "/")==0) to="/.";
    res = link(from+1, to+1);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_chmod(const char *path, mode_t mode)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res = chmod(path+1, mode);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_chown(const char *path, uid_t uid, gid_t gid)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res = lchown(path+1, uid, gid);
    if (res == -1)
        return -errno;

    return 0;
}

int do_truncate(struct ddumb_fh *fh, long long int size)
{// used by other functions to truncate file
    int res;
    struct xstat *xstat=fh->xstat;
    DDFS_LOG_DEBUG("[%lu]++  do_truncate size=0x%llx(%lld) -> 0x%llx(%lld) %s\n", thread_id(), xstat->h.size, xstat->h.size, size, size, fh->filename);

    ddumb_statistic.do_truncate++;

    if (size==xstat->h.size) return 0;
    else if (size<xstat->h.size)
    {   // truncating down

        // I must drop all buffers after the new size

        // When truncating down I don't modify the block on the limit, because
        // it will be filled of zeros when truncating up or rewritten by next write

        long long int block_boundary=size-size%ddfs.c_block_size;

        pthread_mutex_lock_d(&fh->xstat->xstat_lock);
        int retry=1;
        while (retry)
        {   // retry until all collisioning RDONLY buffers have been flushed
            retry=0;
            int i;
            for(i=0; i<fh->xstat->fhs_n; i++)
            {
                struct ddumb_fh *xfh=fh->xstat->fhs[i];
                // DDFS_LOG_DEBUG("[%lu]++  ddumb_write fh=%p xstat=%p xfh=%p I=%d\n", thread_id(), (void*)fh, (void*)fh->xstat, (void*)xfh, i);
                if (xfh->buf_loaded && (xfh->buf_off>block_boundary || xfh->buf_off==size))
                {
                    if (xfh->buf_loaded==DDFS_BUF_RDONLY)
                    {   // the buffer is being flushed, wait for the end
                        retry=1;
                        break;
                    }
                    xfh->buf_loaded=DDFS_BUF_EMPTY;
                }
            }
            if (retry)
            {
                // wait until the buffer has been flushed
                pthread_cond_wait(&fh->xstat->buf_cond, &fh->xstat->xstat_lock);
            }
        }
        pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
    }
    else
    {   // truncating up
        long long int gap=xstat->h.size%ddfs.c_block_size;
//        long long int block_boundary=xstat->h.size-gap;

        if (gap)
        {
            // fill in unused space at end of last block
            DDFS_LOG_DEBUG("[%lu]++  do_truncate write zeroes to end of last block boundary=0x%llx(%lld) gap=0x%llx(%lld) size=%lld %s\n", thread_id(), xstat->h.size-gap, xstat->h.size-gap, gap, gap, ddfs.c_block_size-gap, fh->filename);
            _ddumb_write(NULL, NULL, ddfs.c_block_size-gap, xstat->h.size, fh);
        }

        // truncate fill the file with addr=0x0.....0 (using sparse file technique on unix)
        // and this address is reserved for block"\0......\0"
    }

    // size of the file after truncate
    off_t sz=(size+ddfs.c_block_size-1)/ddfs.c_block_size*ddfs.c_node_size+ddfs.c_file_header_size;

    res=ftruncate(fh->fd, sz);
    if (res==-1)
    {
        DDFS_LOG(LOG_ERR, "do_truncate %lld -> %lld %s: (%s)\n", xstat->h.size, size, fh->filename, strerror(errno));
        return -errno;
    }

    xstat->h.size=size;
    xstat->saved=0;
    return 0;
}

static int ddumb_truncate(const char *path, off_t size)
{
    int res;

    ddumb_statistic.truncate++;

    DDFS_LOG_DEBUG("[%lu]++  ddumb_truncate %s size=0x%llx(%lld)\n", thread_id(), path, (long long int)size, (long long int)size);
    if (strcmp(path, "/")==0) path="/.";

    int fd=open(path+1, O_RDWR);
    if (fd==-1) return -errno;

    struct fuse_file_info fi;
    struct ddumb_fh *fh=ddumb_alloc_fh(&fi, fd, 0, 0, path);
    if (fh==NULL)
    {
        close(fd);
        DDFS_LOG(LOG_ERR, "ddumb_truncate not enough memory for ddumb_alloc_fh(): %s\n", path);
        return -ENOMEM;
    }

    pthread_mutex_lock_d(&fh->lock); // useless because fh cannot be "owned" by another thread
    xzone_lock(fh, size, 0, 'T');
    res=do_truncate(fh, size);
    xzone_unlock(fh);
    ddumb_buffer_flush(fh);
    pthread_mutex_unlock_d(&fh->lock); // useless because fh cannot be "owned" by another thread
    ddumb_free_fh(&fi);

    close(fd);

    return res;
}

static int ddumb_ftruncate(const char *path, off_t size, struct fuse_file_info *fi)
{
    (void) path;
    struct ddumb_fh *fh=ddumb_get_fh(fi);

    pthread_mutex_lock_d(&fh->lock);
    xzone_lock(fh, size, 0, 'T');

    ddumb_statistic.ftruncate++;

    DDFS_LOG_DEBUG("[%lu]++  ddumb_ftruncate %s size=0x%llx(%lld)\n", thread_id(), path, (long long int)size, (long long int)size);

    int res=do_truncate(fh, size);
    if (res==0)
    {   // I think this is important to save the stat now
        res=xstat_save(fh->fd, fh->xstat, fh->filename);
    }

    xzone_unlock(fh);
    pthread_mutex_unlock_d(&fh->lock);

    return res;
}

static int ddumb_utimens(const char *path, const struct timespec ts[2])
{
    int res;
    struct timeval tv[2];

    tv[0].tv_sec = ts[0].tv_sec;
    tv[0].tv_usec = ts[0].tv_nsec / 1000;
    tv[1].tv_sec = ts[1].tv_sec;
    tv[1].tv_usec = ts[1].tv_nsec / 1000;

    if (strcmp(path, "/")==0) path="/.";
    res = utimes(path+1, tv);
    if (res == -1)
        return -errno;

    return 0;
}

static int ddumb_create(const char *path, mode_t mode, struct fuse_file_info *fi)
{
    ddumb_statistic.create++;
//    DDFS_LOG_DEBUG("[%lu]++  ddumb_create %s flags=%o\n", thread_id(), path, fi->flags);

    int special=(0==strncmp(path, SPECIAL_DIR, ddfs.special_dir_len));

    if (special)
    {
        return -EACCES;
    }

    if (strcmp(path, "/")==0) path="/.";

    int flags=fi->flags;
    if (flags & O_WRONLY) flags=(flags^O_WRONLY)|O_RDWR;
    flags=flags&~O_APPEND;

    int fd=open(path+1, flags, mode);
    if (fd==-1)
    {
        DDFS_LOG(LOG_ERR, "ddumb_create open %s flags=%o mode=%o : %s\n", path, fi->flags, mode, strerror(errno));
        return -errno;
    }

    struct ddumb_fh *fh=ddumb_alloc_fh(fi, fd, !(flags&(O_WRONLY|O_RDWR)), 0, path);
    if (fh==NULL)
    {
        close(fd);
        DDFS_LOG(LOG_ERR, "ddumb_create not enough memory for ddumb_alloc_fh(): %s\n", path);
        return -ENOMEM;
    }
    DDFS_LOG_DEBUG("[%lu]++  ddumb_create %s flags=%o fd=%d fh=%p\n", thread_id(), path, fi->flags, fd, (void*)fh);

    return 0;
}

static int ddumb_open(const char *path, struct fuse_file_info *fi)
{
    ddumb_statistic.open++;
    DDFS_LOG_DEBUG("[%lu]++  ddumb_open %s flags=%o rdonly=%d\n", thread_id(), path, fi->flags, !(fi->flags&(O_WRONLY|O_RDWR)));

    int special=(0==strncmp(path, SPECIAL_DIR, ddfs.special_dir_len));

    if (special)
    {
        if ((fi->flags & O_WRONLY) || (fi->flags & O_RDWR)) return -EACCES;
    }

    if (strcmp(path, "/")==0) path="/.";

    int flags=fi->flags;
    if (flags & O_WRONLY) flags=(flags^O_WRONLY)|O_RDWR;
    flags=flags&~O_APPEND;

    int fd=open(path+1, flags);
    if (fd==-1)
    {
        DDFS_LOG(LOG_ERR, "ddumb_open open %s flags=%o : %s\n", path, fi->flags, strerror(errno));
        return -errno;
    }

    struct ddumb_fh *fh=ddumb_alloc_fh(fi, fd, !(flags&(O_WRONLY|O_RDWR)), special, path);
    if (fh==NULL)
    {
        close(fd);
        DDFS_LOG(LOG_ERR, "ddumb_open not enough memory for ddumb_alloc_fh(): %s\n", path);
        return -ENOMEM;
    }

    if (special)
    {
        if (strcmp(path, STATS_FILE)==0 || strcmp(path, STATS0_FILE)==0)
        {
            FILE *file=fopen(path+1, "w");
            ddumb_write_statistic(file);
            sleep(2); // TODO: wait for "attr_timeout", default 1.0s, this will force fuse to reload file size to get true size.
            fclose(file);
            if (strcmp(path, STATS0_FILE)==0) memset(&ddumb_statistic, '\0', sizeof(struct ddumb_statistic));
        }
        else if (strcmp(path, RECLAIM_FILE)==0)
        {
            long long int now();
            FILE *file=fopen(path+1, "w");
            reclaim(file);
            sleep(2); // TODO: wait for "attr_timeout", default 1.0s, this will force fuse to reload file size to get true size.
            fclose(file);

        }
    }

    DDFS_LOG_DEBUG("[%lu]++  ddumb_open %s flags=%o fd=%d fh=%p rdonly=%d %d\n", thread_id(), path, fi->flags, fd, (void*)fh, fh->rdonly, !(flags&(O_WRONLY|O_RDWR)));
/* Be careful because truncate can be called just after open
 * see http://marc.info/?l=fuse-devel&m=129972717512852&w=2
 * This is no more a problem now
 */
    return 0;
}

static int _ddumb_read(const char *path, char *buf, size_t size, off_t offset, struct ddumb_fh *fh)
{
    (void) path;
    int res=0;
    ddumb_statistic.read++;

    DDFS_LOG_DEBUG("[%lu]++  ddumb_read fd=%d offset=0x%llx(%lld) size=0x%llx(%lld) %s\n", thread_id(), fh->fd, (long long int)offset, (long long int)offset, (long long int)size, (long long int)size, path);

    if (offset > fh->xstat->h.size)
    {
        return 0;
    }

    if (offset+size > fh->xstat->h.size)
    {   // don't read more than EOF
        size=fh->xstat->h.size-offset;
    }

    size_t remain=size;
    char *pbuf=buf;
    long long int off=offset;
    long long int gap=off%ddfs.c_block_size;
    long long int block_boundary=off-gap;
    while (remain>0)
    { // split read into multiple aligned block read
        long long int sz=min_int(remain, ddfs.c_block_size-gap);

        res=ddumb_block_read(fh, pbuf, off, sz);
        if (res<0) return res;

        pbuf+=sz;
        remain-=sz;
        off+=sz;

        gap=0;
        block_boundary=off;
    }
    return size;
}

static int ddumb_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
    (void) path;
    int res;
    struct ddumb_fh *fh=ddumb_get_fh(fi);
    if (fh->special)
    {
        res=pread(fh->fd, buf, size, offset);
        if (res==-1) res=-errno;
    }
    else
    {
        xzone_lock(fh, (long long int)offset, (long long int)size, 'R');
        res=_ddumb_read(path, buf, size, offset, fh);
        xzone_unlock(fh);
    }
    return res;
}

static int _ddumb_write(const char *path, const char *buf, size_t size, off_t offset, struct ddumb_fh *fh)
{// be careful _ddumb_write can call itself via do_truncate
    int res=0;
    (void) path;

    ddumb_statistic.write++;
    DDFS_LOG_DEBUG("[%lu]++  ddumb_write fd=%d fh=%p from=0x%llx(%lld) to=0x%llx(%lld) size=0x%llx(%lld) buf_off=%llx xstat_ref=%d %s\n", thread_id(), fh->fd, (void*)fh, (long long int)offset, (long long int)offset, (long long int)(offset+size-1), (long long int)(offset+size-1), (long long int)size, (long long int)size, (long long int)(fh->buf_loaded?fh->buf_off:-1), fh->xstat->ref, path);

    long long int gap=offset%ddfs.c_block_size;
    long long int block_boundary=offset-gap;

    // write after the end of last block ?
    if (block_boundary > fh->xstat->h.size)
    {
        ddumb_statistic.eof_write++;
        DDFS_LOG_DEBUG("[%lu]--  ddumb_write write after last block, file size=0x%llx(%lld) offset=0x%llx(%lld)\n", thread_id(), fh->xstat->h.size, fh->xstat->h.size, (long long int)offset, (long long int)offset);
        res=do_truncate(fh, offset);
        if (res<0) return res;
    }

    long long int remain=size;
    long long int off=offset;
    const char *pbuf=buf;
    while (remain>0)
    { // split write into multiple aligned block writes
        long long int sz=min_int(remain, ddfs.c_block_size-gap);

        // search if pbuf can be written into any available buffers registered in xstat

        int found=0;
        int retry=1; // retry until no RDONLY buffers are in the way
        pthread_mutex_lock_d(&fh->xstat->xstat_lock);
        while (retry)
        {
            retry=0;

            int i;
            for(i=0, found=0; !found && i<fh->xstat->fhs_n; i++)
            {
                struct ddumb_fh *xfh=fh->xstat->fhs[i];
                // DDFS_LOG_DEBUG("[%lu]++  ddumb_write fh=%p xstat=%p xfh=%p I=%d\n", thread_id(), (void*)fh, (void*)fh->xstat, (void*)xfh, i);
                found=(xfh->buf_loaded && xfh->buf_off==block_boundary);
                if (found)
                {
                    if (xfh->buf_loaded==DDFS_BUF_RDONLY)
                    {   // the buffer is being flushed, I need to wait to avoid any race condition
                        retry=1;
                        break;
                    }
                    else
                    {
                        if (xfh!=fh) DDFS_LOG_DEBUG("[%lu]--  ddumb_write write into another fh=%p xfh=%p\n", thread_id(), (void*)fh, (void*)xfh);

                        if (off > fh->xstat->h.size)
                        { // handle write after EOF
                            DDFS_LOG_DEBUG("[%lu]--  ddumb_write write after EOF but inside last block, file size=0x%llx(%lld) offset=0x%llx(%lld)\n", thread_id(), fh->xstat->h.size, fh->xstat->h.size, (long long int)offset, (long long int)offset);
                            memset(xfh->buf+(fh->xstat->h.size-block_boundary), '\0', off-fh->xstat->h.size);
                        }
                        if (pbuf) memcpy(xfh->buf+gap, pbuf, sz);
                        else memset(xfh->buf+gap, '\0', sz);
                    }
                }
            }
            if (retry)
            {   // wait until the buffer has been flushed
                pthread_cond_wait(&fh->xstat->buf_cond, &fh->xstat->xstat_lock);
            }
        }
        pthread_mutex_unlock_d(&fh->xstat->xstat_lock);

        if (!found)
        {

            if (fh->buf_loaded && fh->buf_off!=block_boundary)
            {
                ddumb_statistic.write_save++;
                res=ddumb_buffer_flush(fh);
                if (res<0)
                {
                    return res;
                }
            }

            // fh->buf_loaded is EMPTY now, then no need to lock fh->buf_lock
            // to access fh->buf
            if ((gap!=0 || remain<ddfs.c_block_size))
            {
                DDFS_LOG_DEBUG("[%lu]++  ddumb_write partial write gap=0x%llx(%lld) remain=%lld file_size=%lld loaded=%d %llx!=%llx\n", thread_id(), gap, gap, remain, (long long int)fh->xstat->h.size, fh->buf_loaded, (long long int)fh->buf_off, (long long int)(off-gap));
                if (block_boundary < fh->xstat->h.size)
                {
                    ddumb_statistic.read_before_write++;
                    DDFS_LOG_DEBUG("[%lu]--  ddumb_write read_before_write fd=%d fh=%p, offset=0x%llx(%lld)\n", thread_id(), fh->fd, (void*)fh, (long long int)block_boundary, (long long int)block_boundary);
                    int psize=ddfs.c_block_size;
                    if (block_boundary+psize>fh->xstat->h.size)
                    {
                        psize=fh->xstat->h.size-block_boundary;
                    }
                    res=ddumb_block_read(fh, fh->buf, block_boundary, psize);
                    if (res<0)
                    {
                        return res;
                    }
                }

                if (off > fh->xstat->h.size)
                {   // if end of block > file size, fill block with zeros from end of file up to start of offset
                    assert(block_boundary+ddfs.c_block_size-fh->xstat->h.size<=ddfs.c_block_size);
                    memset(fh->buf+(fh->xstat->h.size-block_boundary), '\0', off-fh->xstat->h.size);
                }
            }

            if (pbuf) memcpy(fh->buf+gap, pbuf, sz);
            else memset(fh->buf+gap, '\0', sz);
            fh->buf_off=block_boundary;
            pthread_mutex_lock_d(&fh->xstat->xstat_lock);
            fh->buf_loaded=DDFS_BUF_RDWR;
            pthread_mutex_unlock_d(&fh->xstat->xstat_lock);
        }

        res=0;
        if (fh->buf_loaded && fh->buf_off==block_boundary && gap+sz==ddfs.c_block_size)
        {
            res=ddumb_buffer_flush(fh);
        }
        if (res<0) return res;

        if (pbuf) pbuf+=sz;
        remain-=sz;
        off+=sz;
        gap=0;
        block_boundary+=ddfs.c_block_size;

        // update fh->xstat->h.size
        // TODO: maybe lock something here
        if (off > fh->xstat->h.size)
        {
            fh->xstat->h.size=off;
            fh->xstat->saved=0;
        }
    }
    return size;
}

static int ddumb_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
    struct ddumb_fh *fh=ddumb_get_fh(fi);
    pthread_mutex_lock_d(&fh->lock);

    xzone_lock(fh, (long long int)offset, (long long int)size, 'W');
    int res=_ddumb_write(path, buf, size, offset, fh);
    xzone_unlock(fh);

    pthread_mutex_unlock_d(&fh->lock);
    return res;
}

static int ddumb_statfs(const char *path, struct statvfs *stbuf)
{
    int res;

    if (strcmp(path, "/")==0) path="/.";
    res=statvfs(path+1, stbuf);
    if (res==-1) return -errno;

    long long int s, u;
    pthread_spin_lock(&ifile_spinlock);
    bit_array_count(&ddfs.ba_usedblocks, &s, &u);
    pthread_spin_unlock(&ifile_spinlock);
    stbuf->f_bfree=stbuf->f_bfree*(stbuf->f_frsize/ddfs.c_addr_size);
    if (u<stbuf->f_bfree) stbuf->f_bfree=u;
    stbuf->f_bavail=stbuf->f_bavail*(stbuf->f_frsize/ddfs.c_addr_size);
    if (u<stbuf->f_bavail) stbuf->f_bavail=u;
    stbuf->f_bsize=ddfs.c_block_size;
    stbuf->f_frsize=ddfs.c_block_size;
    stbuf->f_blocks=ddfs.c_block_count;
    return 0;
}


static void _ddumb_flush(struct ddumb_fh *fh)
{
    // be careful not to flush special file that are RDONLY and don't have buffer nor buf_lock
    if (fh->buf)
    {
        pthread_mutex_lock_d(&fh->lock);
        ddumb_buffer_flush(fh);
        pthread_mutex_unlock_d(&fh->lock);
    }

    if (ddumb_param.pool)
    {   // wait for the writers of the pool
        pthread_mutex_lock(&writer_pool_mutex);
        while (fh->pool_writer>0) pthread_cond_wait(&writer_pool_cond_done, &writer_pool_mutex);
        pthread_mutex_unlock(&writer_pool_mutex);
    }
    if (fh->xstat && !fh->xstat->saved) xstat_save_fh(fh);

}

static int ddumb_flush(const char *path, struct fuse_file_info *fi)
{
    int res;

    (void) path;
    /* This is called from every close on an open file (not sure if even read only),
       so call the close on the underlying filesystem. But since flush may be
       called multiple times for an open file, this must not really
       close the file.  This is important if used on a network
       filesystem like NFS which flush the data/metadata on close().
       this is not called for the last close(), release() is called instead  */
    struct ddumb_fh *fh=ddumb_get_fh(fi);
    DDFS_LOG_DEBUG("[%lu]++  ddumb_flush fd=%d fh=%p %s\n", thread_id(), fh->fd, (void*)fh, path);
    ddumb_statistic.flush++;

    _ddumb_flush(fh);

    int fd=dup(fh->fd);
    res=close(fd);
    if (res==-1)
    {
        DDFS_LOG(LOG_ERR, "ddumb_flush close(dup) %s (%s)\n", fh->filename, strerror(errno));
        return -errno;
    }

    return 0;
}

static int ddumb_release(const char *path, struct fuse_file_info *fi)
{
    int res=0;
    (void) path;
    ddumb_statistic.release++;
    struct ddumb_fh *fh=ddumb_get_fh(fi);
    DDFS_LOG_DEBUG("[%lu]++  ddumb_release fd=%d fh=%p buf_loaded=%d xstat->saved=%d %s\n", thread_id(), fh->fd, (void*)fh, fh->buf?fh->buf_loaded:99, fh->xstat?fh->xstat->saved:99, path);

    _ddumb_flush(fh);

    assert(!fh || !fh->buf || !fh->buf_loaded);
    assert(!fh->xstat || fh->xstat->saved || fh->rdonly);

    int fd=fh->fd;
    res=ddumb_free_fh(fi);
    close(fd);
    return res;
}

static int ddumb_fsync(const char *path, int isdatasync, struct fuse_file_info *fi)
{
    int res;
    (void) path;

    struct ddumb_fh *fh=ddumb_get_fh(fi);
    DDFS_LOG_DEBUG("[%lu]++  ddumb_fsync fd=%d fh=%p %s\n", thread_id(), fh->fd, (void*)fh, path);
    ddumb_statistic.fsync++;

    _ddumb_flush(fh);

#ifndef HAVE_FDATASYNC
    (void) isdatasync;
#else
    if (isdatasync)
        res = fdatasync(fh->fd);
    else
#endif
        res = fsync(fh->fd);

    if (res==-1) return -errno;
    res=fsync(ddfs.bfile);
    if (res==-1) return -errno;
    res=fsync(ddfs.ifile);
    if (res==-1) return -errno;

    return 0;
}

#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int ddumb_setxattr(const char *path, const char *name, const char *value, size_t size, int flags)
{
    if (strcmp(path, "/")==0) path="/.";
    int res=lsetxattr(path+1, name, value, size, flags);
    if (res==-1) return -errno;
    return 0;
}

static int ddumb_getxattr(const char *path, const char *name, char *value, size_t size)
{
    if (strcmp(path, "/")==0) path="/.";
    int res=lgetxattr(path+1, name, value, size);
    if (res==-1) return -errno;
    return res;
}

static int ddumb_listxattr(const char *path, char *list, size_t size)
{
    if (strcmp(path, "/")==0) path="/.";
    int res=llistxattr(path+1, list, size);
    if (res==-1) return -errno;
    return res;
}

static int ddumb_removexattr(const char *path, const char *name)
{
    if (strcmp(path, "/")==0) path="/.";
    int res=lremovexattr(path+1, name);
    if (res==-1)return -errno;
    return 0;
}
#endif /* HAVE_SETXATTR */

static int ddumb_lock(const char *path, struct fuse_file_info *fi, int cmd, struct flock *lock)
{
    (void) path;

    struct ddumb_fh *fh=ddumb_get_fh(fi);
    int res=ulockmgr_op(fh->fd, cmd, lock, &fi->lock_owner, sizeof(fi->lock_owner));
    return res;
}

void *ddumbfs_background(void *ptr)
{
    struct timeval now;
    struct timespec timeout;

    pthread_mutex_lock(&ddumb_background_mutex);

    while (!ddumbfs_background_terminate)
    {
        gettimeofday(&now, NULL);
        timeout.tv_sec=now.tv_sec+60;
        timeout.tv_nsec=now.tv_usec*1000;
        int res=pthread_cond_timedwait(&ddumb_background_cond, &ddumb_background_mutex, &timeout);

        pthread_spin_lock(&ifile_spinlock);
        if (res==ETIMEDOUT) ddumbfs_save_usedblocks(1000);
        else ddumbfs_save_usedblocks(0);
        pthread_spin_unlock(&ifile_spinlock);
    }
    pthread_mutex_unlock(&ddumb_background_mutex);
    pthread_exit(NULL);
}

static void* ddumb_init(struct fuse_conn_info *conn)
{
    chdir(ddfs.rdir);
    if (ddfs.auto_fsck || ddumb_param.check_at_start)
    {
        DDFS_LOG(LOG_INFO, "check filesystem %s\n", ddumb_param.parent);
        L_SYS(LOG_INFO, "check filesystem %s\n", ddumb_param.parent);
        ddfs_fsck(DDFS_NORELAXED, DDFS_NOVERBOSE, DDFS_NOPROGRESS);
    }

    if (ddfs_lock(".autofsck"))
    {
        DDFS_LOG(LOG_ERR, "cannot create .autofsck file, continue: %s\n", strerror(errno));
    }

    DDFS_LOG(LOG_INFO, "filesystem %s mounted\n", ddumb_param.parent);
    L_SYS(LOG_INFO, "filesystem %s mounted\n", ddumb_param.parent);

    // ddfs_save_usedblocks(); // will be done by ddumbfs_background at startup

    pthread_create(&ddumbfs_background_pthread, NULL, ddumbfs_background, NULL);

    int res;
    if (ddumb_param.pool)
    {
        res=init_writer_pool();
        if (res);
    }

    return NULL;
}

void free_xstat_node(void *nodep)
{  // used by ddumb_destroy to report unclosed file
    DDFS_LOG(LOG_ERR, "free_node %p (should not append when all file are closed before unmount)\n", nodep)
}

static void ddumb_destroy(void* nothing)
{ // called just after umount

    // display statistics
    ddumb_write_statistic(stderr);

    // destroy xstat to be sure no one was left behind (except still open file)
    tdestroy(xstat_root.root, free_xstat_node);

    pthread_mutex_lock(&ddumb_background_mutex);
    ddumbfs_background_terminate=1;
    pthread_cond_signal(&ddumb_background_cond);
    pthread_mutex_unlock(&ddumb_background_mutex);

    pthread_join(ddumbfs_background_pthread, NULL);

    ddfs_close();

    // ok cleanly unmounted
    ddfs_unlock(".autofsck");
    L_SYS(LOG_INFO, "filesystem %s unmounted\n", ddumb_param.parent);
    DDFS_LOG(LOG_INFO, "filesystem %s unmounted\n", ddumb_param.parent);
}

static struct fuse_operations ddumb_ops = {
    .getattr        = ddumb_getattr,
    .fgetattr       = ddumb_fgetattr,
    .access         = ddumb_access,
    .readlink       = ddumb_readlink,
    .opendir        = ddumb_opendir,
    .readdir        = ddumb_readdir,
    .releasedir     = ddumb_releasedir,
    .mknod          = ddumb_mknod,
    .mkdir          = ddumb_mkdir,
    .symlink        = ddumb_symlink,
    .unlink         = ddumb_unlink,
    .rmdir          = ddumb_rmdir,
    .rename         = ddumb_rename,
    .link           = ddumb_link,
    .chmod          = ddumb_chmod,
    .chown          = ddumb_chown,
    .ftruncate      = ddumb_ftruncate,
    .truncate       = ddumb_truncate,
    .utimens        = ddumb_utimens,
    .create         = ddumb_create,
    .open           = ddumb_open,
    .read           = ddumb_read,
    .write          = ddumb_write,
    .statfs         = ddumb_statfs,
    .flush          = ddumb_flush,
    .release        = ddumb_release,
    .fsync          = ddumb_fsync,
#ifdef HAVE_SETXATTR
    .setxattr       = ddumb_setxattr,
    .getxattr       = ddumb_getxattr,
    .listxattr      = ddumb_listxattr,
    .removexattr    = ddumb_removexattr,
#endif
    .lock           = ddumb_lock,
    .init           = ddumb_init,
    .destroy        = ddumb_destroy,
#if FUSE_VERSION >= 28
    .flag_nullpath_ok = 1,
#endif
};


enum {
     KEY_HELP,
     KEY_VERSION,
     KEY_DEBUG,
};


#define DDUMB_OPT(t, p, v) { t, offsetof(struct_ddumb_param, p), v }

static const struct fuse_opt ddumb_opts[] = {
        DDUMB_OPT("parent=%s", parent, 0),
        DDUMB_OPT("pool=%i", pool, 0),
        DDUMB_OPT("nopool", pool, -1),
        DDUMB_OPT("check", check_at_start, 1),
        DDUMB_OPT("lock_index", lock_index, 1),
        DDUMB_OPT("nolock_index", lock_index, 0),

        DDUMB_OPT("dio", direct_io, 1),
        DDUMB_OPT("nodio", direct_io, 0),

        FUSE_OPT_KEY("-d", KEY_DEBUG),

        FUSE_OPT_KEY("-h", KEY_HELP),
        FUSE_OPT_KEY("--help", KEY_HELP),

        FUSE_OPT_KEY("-V", KEY_VERSION),
        FUSE_OPT_KEY("--version", KEY_VERSION),

        FUSE_OPT_END
};

static int ddumb_opt_proc(void *data, const char *arg, int key, struct fuse_args *outargs)
{
//    struct ddumb_param *param = data;

    (void) arg;
    (void) outargs;

//    fprintf(stderr,"ddumbr_opt_proc key=%d arg=%s\n", key, arg);

    switch (key) {
    case KEY_HELP:
            fuse_opt_add_arg(outargs, "-h");
            fuse_main(outargs->argc, outargs->argv, &ddumb_ops, NULL);
            fprintf(stderr,
//                    "usage: %s mountpoint [options]\n"
                    "\n"
                    "ddumbfs options:\n"
                    "\n"
                    "    -o parent=DIR      directory to mount as a ddumbfs\n"
                    "    -o pool=NUM        number of writer in the pool, 0=disable, <0=CPUs*NUM/10, >0=CPUs\n"
                    "    -o [no]lock_index  lock index into memory (default on)\n"
                    "    -o [no]dio         use direct_io for internal access (default on)\n"
                    "    -o check           force filesystem check at startup\n"
                    "\n"
//                    , outargs->argv[0]
                                    );
            exit(1);

    case KEY_VERSION:
            fprintf(stderr, "ddumbfs version %s\n", VERSION);
            fuse_opt_add_arg(outargs, "--version");
            fuse_main(outargs->argc, outargs->argv, &ddumb_ops, NULL);
            exit(0);

    case KEY_DEBUG:
            fprintf(stderr, "debug ON\n");
            ddfs_debug=1;
            return 1;
    }
    return 1;
}

int main(int argc, char *argv[])
{
    struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
    char line[1024];
    long long int i;
    char ch;

    umask(0);
    // reset all stats to 0
    memset(&ddumb_statistic, 0x00, sizeof(struct ddumb_statistic));

    if (fuse_opt_parse(&args, &ddumb_param, ddumb_opts, ddumb_opt_proc)==1)
    {
        // error  parsing options
        return -1;
    }

    if (!ddumb_param.parent) {
        fprintf(stderr, "ERROR: parent directory is missing\n");
        return 1;
    }

    char pdir[FILENAME_MAX];

    if (ddfs_find_parent(ddumb_param.parent, pdir))
    {
        fprintf(stderr, "Not a valid ddumbfs filesystem parent directory: %s\n", argv[optind]);
        return -1;
    }
    strcpy(ddumb_param.parent, pdir); // I can copy back pdir because len(pdir)<=len(ddumb_param.parent)

    if (is_mounted(ddumb_param.parent))
    {
        fprintf(stderr, "filesystem already mounted: %s\n", ddumb_param.parent);
        return 1;
    }

    //
    // load config file
    //
    if (ddfs_loadcfg(ddumb_param.parent, stderr))
    {
        return 1;
    }

    if (ddfs.rebuild_fsck)
    {
        fprintf(stderr, "Last file system rebuild didn't finished, 'rebuild' your file system before mounting it !\n");
        return 1 ;
    }

    fprintf(stderr,"HASH %s\n", ddfs.c_hash);

    if (ddfs.direct_io)
    {
        fprintf(stderr,"DIRECT_IO enable\n");
    }

#ifdef ULTRAFAST
    fprintf(stderr,"!!!!!!!!!! ULTRAFAST enable !!!!!!!!!!!!!!!\n");
#else

#endif

    //
    // init ddfs
    //
    if (ddfs_init(DDFS_NOFORCE, ddumb_param.direct_io, ddumb_param.lock_index, stderr))
    {
        exit(1);
    }

    if (ddfs_debug)
    {
        ddfs_logger->level=LOG_DEBUG;
        ddfs_logger->handlers[0].level=LOG_DEBUG;
    }

    //
    // init some aux bit array
    //
    bit_array_init(&ba_found_in_files, ddfs.c_block_count, 0);

    // init xstat root
    xstat_root.root=NULL;
    pthread_mutex_init(&xstat_root.mutex, NULL);
    xstat_root.xstat=NULL;

    pthread_spin_init(&reclaim_spinlock, 0);
    pthread_spin_init(&ifile_spinlock, 0);

    // force all index block to be read/loaded at statup
    for (i=0; i<ddfs.c_node_block_count; i++) ch=ddfs.nodes[i*ddfs.c_index_block_size];

    chdir(ddfs.rdir);
    if (!isdir(".ddumbfs"))
    {
        char cmdbuf[1024];
        fprintf(stderr, "ERROR system directory .ddumbfs not found in %s!\n", getcwd(cmdbuf, sizeof(cmdbuf)));
        exit(1);
    }

    //
    // writers pool
    //
    if (ddumb_param.pool<0)
    {   // pool=auto, search number of cpu
        int count=0;
        char *cpuinfo="/proc/cpuinfo";
        FILE *cpufile=fopen(cpuinfo, "r");
        if (!cpufile)
        {
            perror(cpuinfo);
            return 1;
        }

        while (fgets(line, sizeof(line), cpufile))
        {
            char *key=trim(strtok(line, ":"));
            //char *value=trim(strtok(NULL, "\n"));

            if (strcmp("processor", key)==0)
            {
                count++;
            }
        }
        ddumb_param.pool=-0.1*ddumb_param.pool*count;
        if (ddumb_param.pool==0) ddumb_param.pool=1;
    }

    if (ddumb_param.pool)
    {
        fprintf(stderr, "writer pool: %d cpus\n", ddumb_param.pool);
        writer_pool_threads=malloc(ddumb_param.pool*sizeof(pthread_t));
        if (writer_pool_threads==NULL)
        {
            fprintf(stderr, "ERROR cannot allocate writer pool table[%d]!\n", ddumb_param.pool);
            exit(1);
        }
    }
    else
    {
        fprintf(stderr, "writer pool: no\n");
    }


//    snfprintf(stderr,subdir_root, sizeof(subdir_root), "-omodules=subdir,subdir=/%s/%s", param.root, TOP_DIR);
//    fuse_opt_add_arg(&args, subdir_root);

    /* be careful, fuse can generate multiple simultaneous read. ddumbfs_read()
     * in direct_io mode use a unique aligned buffer for its read, it must be protected
     * to avoid simultaneous access, or "-osync_read" could be used to disable this feature
     * fuse_opt_add_arg(&args, "-osync_read");
     */

    snprintf(line, sizeof(line), "-osubtype=ddumbfs,fsname=%s", ddumb_param.parent);
    fuse_opt_add_arg(&args, line);

    return fuse_main(args.argc, args.argv, &ddumb_ops, NULL);
}
